use std::time::SystemTime;

use wasmi::core::Trap;

use super::MinimalPreview1;

pub struct StockWasi {

}
impl StockWasi {
    pub fn new() -> Self {
        Self { }
    }
}

// https://github.com/WebAssembly/wasi-libc/blob/main/libc-bottom-half/headers/public/wasi/api.h#L118
const BAD_DESCRIPTOR: i32 = 8;
const FUNCTION_NOT_IMPLEMENTED: i32 = 52;

// Implemented based on docs here
// https://wasix.org/docs/api-reference/wasi/fd_allocate
// https://github.com/WebAssembly/WASI/blob/main/legacy/preview1/witx/wasi_snapshot_preview1.witx
impl MinimalPreview1 for StockWasi {
    fn args_get(&mut self, _memory: &mut [u8], _argv: i32, _argv_buf: i32) -> super::MPResult<i32> {
        Ok(0)
    }

    fn args_sizes_get(&mut self, memory: &mut [u8], argc: i32, argv_buf_size: i32) -> super::MPResult<i32> {
        write_u32(memory, argc, 0);
        write_u32(memory, argv_buf_size, 0);
        Ok(0)
    }

    fn environ_get(&mut self, _memory: &mut [u8], _environ: i32, _environ_buf: i32) -> super::MPResult<i32> {
        Ok(0)
    }

    fn environ_sizes_get(&mut self, memory: &mut [u8], environc: i32, environ_buf_size: i32) -> super::MPResult<i32> {
        write_u32(memory, environc, 0);
        write_u32(memory, environ_buf_size, 0);
        Ok(0)
    }

    fn clock_res_get(&mut self, memory: &mut [u8], _id: i32, outptr: i32) -> super::MPResult<i32> {
        write_u32(memory, outptr, 1000);
        Ok(0)
    }

    fn clock_time_get(&mut self, memory: &mut [u8], id: i32, precision: i64, outptr: i32) -> super::MPResult<i32> {
        println!("clock_time_get: {} {} {}", id, precision, outptr);
        // TODO: Actually fetch time
        let duration = 
            SystemTime::now().duration_since(SystemTime::UNIX_EPOCH)
            .expect("time must be post-epoch");

        write_u64(memory, outptr, duration.as_nanos() as u64);
        Ok(0)
    }

    fn fd_advise(&mut self, _memory: &mut [u8], _fd: i32, _offset: i64, _len: i64, _advice: i32) -> super::MPResult<i32> {
        Ok(FUNCTION_NOT_IMPLEMENTED)
    }

    fn fd_allocate(&mut self, _memory: &mut [u8], _fd: i32, _offset: i64, _len: i64) -> super::MPResult<i32> {
        Ok(FUNCTION_NOT_IMPLEMENTED)
    }

    fn fd_close(&mut self, _memory: &mut [u8], _fd: i32) -> super::MPResult<i32> {
        Ok(FUNCTION_NOT_IMPLEMENTED)
    }

    fn fd_datasync(&mut self, _memory: &mut [u8], _fd: i32) -> super::MPResult<i32> {
        Ok(FUNCTION_NOT_IMPLEMENTED)
    }

    fn fd_fdstat_get(&mut self, _memory: &mut [u8], _fd: i32, _offset0: i32) -> super::MPResult<i32> {
        Ok(FUNCTION_NOT_IMPLEMENTED)
    }

    fn fd_fdstat_set_flags(&mut self, _memory: &mut [u8], _fd: i32, _flags: i32) -> super::MPResult<i32> {
        Ok(FUNCTION_NOT_IMPLEMENTED)
    }

    fn fd_fdstat_set_rights(&mut self, _memory: &mut [u8], _fd: i32, _fs_rights_base: i64, _fs_rights_inheriting: i64) -> super::MPResult<i32> {
        Ok(FUNCTION_NOT_IMPLEMENTED)
    }

    fn fd_filestat_get(&mut self, _memory: &mut [u8], _fd: i32, _offset0: i32) -> super::MPResult<i32> {
        Ok(FUNCTION_NOT_IMPLEMENTED)
    }

    fn fd_filestat_set_size(&mut self, _memory: &mut [u8], _fd: i32, _size: i64) -> super::MPResult<i32> {
        Ok(FUNCTION_NOT_IMPLEMENTED)
    }

    fn fd_filestat_set_times(&mut self, _memory: &mut [u8], _fd: i32, _atim: i64, _mtim: i64, _fst_flags: i32) -> super::MPResult<i32> {
        Ok(FUNCTION_NOT_IMPLEMENTED)
    }

    fn fd_pread(&mut self, _memory: &mut [u8], _fd: i32, _iov_buf: i32, _iov_buf_len: i32, _offset: i64, _offset0: i32) -> super::MPResult<i32> {
        Ok(FUNCTION_NOT_IMPLEMENTED)
    }

    fn fd_prestat_get(&mut self, _memory: &mut [u8], fd: i32, _offset0: i32) -> super::MPResult<i32> {
        // for some reason TinyGo asks about fd 3 immediately
        if fd == 3 {
            return Ok(BAD_DESCRIPTOR);
        }
        Ok(FUNCTION_NOT_IMPLEMENTED)
    }

    fn fd_prestat_dir_name(&mut self, _memory: &mut [u8], _fd: i32, _path: i32, _path_len: i32) -> super::MPResult<i32> {
        Ok(FUNCTION_NOT_IMPLEMENTED)
    }

    fn fd_pwrite(&mut self, _memory: &mut [u8], _fd: i32, _ciov_buf: i32, _ciov_buf_len: i32, _offset: i64, _offset0: i32) -> super::MPResult<i32> {
        Ok(FUNCTION_NOT_IMPLEMENTED)
    }

    fn fd_read(&mut self, _memory: &mut [u8], _fd: i32, _iov_buf: i32, _iov_buf_len: i32, _offset1: i32) -> super::MPResult<i32> {
        Ok(FUNCTION_NOT_IMPLEMENTED)
    }

    fn fd_readdir(&mut self, _memory: &mut [u8], _fd: i32, _buf: i32, _buf_len: i32, _cookie: i64, _offset0: i32) -> super::MPResult<i32> {
        Ok(FUNCTION_NOT_IMPLEMENTED)
    }

    fn fd_renumber(&mut self, _memory: &mut [u8], _fd: i32, _to: i32) -> super::MPResult<i32> {
        Ok(FUNCTION_NOT_IMPLEMENTED)
    }

    fn fd_seek(&mut self, _memory: &mut [u8], _fd: i32, _offset: i64, _whence: i32, _offset0: i32) -> super::MPResult<i32> {
        Ok(FUNCTION_NOT_IMPLEMENTED)
    }

    fn fd_sync(&mut self, _memory: &mut [u8], _fd: i32) -> super::MPResult<i32> {
        Ok(FUNCTION_NOT_IMPLEMENTED)
    }

    fn fd_tell(&mut self, _memory: &mut [u8], _fd: i32, _offset0: i32) -> super::MPResult<i32> {
        Ok(FUNCTION_NOT_IMPLEMENTED)
    }

    fn fd_write(&mut self, memory: &mut [u8], fd: i32, ciov_buf: i32, ciov_buf_len: i32, n_written_ptr: i32) -> super::MPResult<i32> {
        let mut out = vec![];
        let mut n_written = 0;
        for i in 0..ciov_buf_len {
            let iov_ptr = ciov_buf + i * 8;
            let ptr = read_u32(memory, iov_ptr);
            let len = read_u32(memory, iov_ptr + 4);
            n_written += len;

            for b in 0..len {
                out.push(memory[ptr as usize + b as usize]);
            }
        }

        write_u32(memory, n_written_ptr, n_written);

        if fd == 1 {
            // stdout
            println!("Written to stdout: {}", String::from_utf8_lossy(&out));
        }
        else if fd == 2 {
            // stderr
            println!("Written to stderr: {}", String::from_utf8_lossy(&out));
        } else {
            println!("Written to invalid FD {}: {}", fd, String::from_utf8_lossy(&out));
        }

        Ok(0)
    }

    fn path_create_directory(&mut self, _memory: &mut [u8], _fd: i32, _offset: i32, _length: i32) -> super::MPResult<i32> {
        Ok(FUNCTION_NOT_IMPLEMENTED)
    }

    fn path_filestat_get(&mut self, _memory: &mut [u8], _fd: i32, _flags: i32, _offset: i32, _length: i32, _offset0: i32) -> super::MPResult<i32> {
        Ok(FUNCTION_NOT_IMPLEMENTED)
    }

    fn path_filestat_set_times(
        &mut self,
        _memory: &mut [u8],
        _fd: i32,
        _flags: i32,
        _offset: i32,
        _length: i32,
        _atim: i64,
        _mtim: i64,
        _fst_flags: i32,
    ) -> super::MPResult<i32> {
        Ok(0)
    }

    fn path_link(
        &mut self,
        _memory: &mut [u8],
        _old_fd: i32,
        _old_flags: i32,
        _old_offset: i32,
        _old_length: i32,
        _new_fd: i32,
        _new_offset: i32,
        _new_length: i32,
    ) -> super::MPResult<i32> {
        Ok(0)
    }

    fn path_open(
        &mut self,
        _memory: &mut [u8],
        _fd: i32,
        _dirflags: i32,
        _offset: i32,
        _length: i32,
        _oflags: i32,
        _fs_rights_base: i64,
        _fdflags: i64,
        _fs_rights_inheriting: i32,
        _offset0: i32,
    ) -> super::MPResult<i32> {
        Ok(0)
    }

    fn path_readlink(
        &mut self,
        _memory: &mut [u8],
        _fd: i32,
        _offset: i32,
        _length: i32,
        _buf: i32,
        _buf_len: i32,
        _offset0: i32,
    ) -> super::MPResult<i32> {
        Ok(0)
    }

    fn path_remove_directory(
        &mut self,
        _memory: &mut [u8],
        _fd: i32, _offset: i32, _length: i32
    ) -> super::MPResult<i32> {
        Ok(0)
    }

    fn path_rename(
        &mut self,
        _memory: &mut [u8],
        _fd: i32,
        _old_offset: i32,
        _old_length: i32,
        _new_fd: i32,
        _new_offset: i32,
        _new_length: i32,
    ) -> super::MPResult<i32> {
        Ok(0)
    }

    fn path_symlink(
        &mut self,
        _memory: &mut [u8],
        _old_offset: i32,
        _old_length: i32,
        _fd: i32,
        _new_offset: i32,
        _new_length: i32,
    ) -> super::MPResult<i32> {
        Ok(0)
    }

    fn path_unlink_file(&mut self, _memory: &mut [u8], _fd: i32, _offset: i32, _length: i32) -> super::MPResult<i32> {
        Ok(0)
    }

    fn poll_oneoff(&mut self, _memory: &mut [u8], _in_: i32, _out: i32, _nsubscriptions: i32, _offset0: i32) -> super::MPResult<i32> {
        Ok(0)
    }

    fn proc_exit(&mut self, _memory: &mut [u8], rval: i32) -> super::MPResult<()> {
        Err(Trap::i32_exit(rval))
    }

    fn proc_raise(&mut self, _memory: &mut [u8], _sig: i32) -> super::MPResult<i32> {
        // TODO: Actually implement
        Ok(0)
    }

    fn sched_yield(&mut self, _memory: &mut [u8]) -> super::MPResult<i32> {
        Ok(0)
    }

    fn random_get(&mut self, _memory: &mut [u8], _buf: i32, _buf_len: i32) -> super::MPResult<i32> {
        // TODO: Actually implement
        Ok(0)
    }

    fn sock_accept(&mut self, _memory: &mut [u8], _fd: i32, _flags: i32, _offset0: i32) -> super::MPResult<i32> {
        Ok(0)
    }

    fn sock_recv(
        &mut self,
        _memory: &mut [u8],
        _fd: i32,
        _iov_buf: i32,
        _iov_buf_len: i32,
        _ri_flags: i32,
        _offset0: i32,
        _offset1: i32,
    ) -> super::MPResult<i32> {
        Ok(0)
    }

    fn sock_send(
        &mut self,
        _memory: &mut [u8],
        _fd: i32, _ciov_buf: i32, _ciov_buf_len: i32, _si_flags: i32, _offset0: i32) -> super::MPResult<i32> {
        Ok(0)
    }

    fn sock_shutdown(
        &mut self,
        _memory: &mut [u8],
        _fd: i32, _how: i32) -> super::MPResult<i32> {
        Ok(0)
    }
}

fn read_u32(memory: &[u8], ptr: i32) -> u32 {
    let ptr = ptr as usize;
    let slice = &memory[ptr..ptr + 4];
    let mut ary = [0; 4];
    ary.copy_from_slice(slice);
    return u32::from_le_bytes(ary)
}

fn write_u32(memory: &mut [u8], ptr: i32, value: u32) {
    let ptr = ptr as usize;
    (&mut memory[ptr..ptr + 4]).copy_from_slice(&value.to_le_bytes());
}

fn write_u64(memory: &mut [u8], ptr: i32, value: u64) {
    let ptr = ptr as usize;
    (&mut memory[ptr..ptr + 8]).copy_from_slice(&value.to_le_bytes());
}